#if !defined (_DECO_H_)
#define _DECO_H_

#define DECO_ERROR_OK 0
#define DECO_ERROR_MEMORY 1
#define DECO_ERROR_NO_FEATURE 2
#define DECO_ERROR_MAX_DEPTH 3
#define DECO_ERROR_GAS_NUMBER 4
#define DECO_ERROR_SETPOINT_NUMBER 5

unsigned int deco_errno = 0;

/*!
 * Dive profile segment.
 */
typedef struct {
    /*! Planned dive bottom time. */
    unsigned char depth;

    /*! Planned dive time at depth. */
    unsigned int time;
} deco_profile_segment_t;

/*!
 * Dive plan profile information.
 */
typedef struct {
    /*! Amount of dive profile segments. */
    unsigned char n_segments;

    /*! Dive profile segments. */
    deco_profile_segment_t *segments;
} deco_profile_t;

/*!
 * CCR setpoint information.
 */
typedef struct {
    /*! Depth at which setpoint is on. */
    unsigned char depth;

    /*! The ppO2 of the setpoint. */
    float ppO2;
} deco_setpoint_t;

/*!
 * Gas mix information.
 */
typedef struct {
    /*! Depth at which gas is to be used. */
    unsigned char depth;

    /*! O2 perecentage. */
    unsigned char o2;

    /*! Helium percentage. */
    unsigned char he;
} deco_gas_mix_t;

/*!
 * Dive plan configuration.
 */
typedef struct {
    /*! Gradient factor low. */
    unsigned char gf_low;

    /*! Gradient factor high. */
    unsigned char gf_high;

    /*! Surface pressure before dive in mbar. */
    unsigned short pressure;

    /*! Last deco stop, i.e. 3m or 6m. */
    unsigned char last_stop;

    /*! CCR setpoint number. */
    unsigned char n_setpoints;

    /*! CCR setpoint information. */
    deco_setpoint_t *setpoints;

    /*! Amount of gas mixes. */
    unsigned char n_gas_mixes;

    /*! List of gas mixes. */
    deco_gas_mix_t *gas_list;
} deco_conf_t;

/*!
 * Deco stop information.
 */
typedef struct {
    unsigned char depth;
    unsigned short time;
} deco_stop_t;

/*!
 * Decompression information.
 */
typedef struct {
    /*! NDL in minutes (from the start of a dive). */
    unsigned short ndl;

    /*! Amount of deco stops.*/
    unsigned char n_stops;

    /*! List of deco stops. */
    deco_stop_t *stops;
} deco_info_t;

/*!
 * Calculate decompression information.
 */
deco_info_t *deco_calculate(deco_profile_t *profile, deco_conf_t *conf);

/*!
 * Add dive profile segment to dive profile.
 */
int deco_segment_add(deco_profile_t *profile, unsigned char depth,
        unsigned int time);

/*!
 * Add gas mix to dive configuration.
 */
int deco_gas_add(deco_conf_t *conf, unsigned char depth, unsigned char o2,
        unsigned char he);

/*!
 * Add setpoint to dive configuration.
 */
int deco_setpoint_add(deco_conf_t *conf, unsigned char depth, float ppO2);

/*!
 * Create new dive configuration with air in gas list.
 */
deco_conf_t *deco_conf_new();

/*!
 * Set altitude for decompression calculations.
 *
 * The pressure of the deco configuration structure is updated.
 */
void deco_set_altitude(deco_conf_t *conf, unsigned short altitude);

/*!
 * Free libdeco memory.
 */
void deco_free(deco_profile_t *profile, deco_conf_t *conf, deco_info_t *info);

/*!
 * Return libdeco error message.
 */
char *deco_error(const int errno);

#endif /* _DECO_H_ */
