#
# DecoMecha - diving decompression library.
#
# Copyright (C) 2013 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Shallower dive sanity unit tests. 
"""

from decomecha import DecoEngine

from .util import TestCase


class ShallowerDecoDiveTestCase(TestCase):
    """
    Shallower depth sanity unit tests for decompression dives.
    """
    def test_oc_air(self):
        """
        Test shallower OC dive on air
        """
        d1 = DecoEngine()
        d1.add_segment(45, 25)
        d1.add_gas(0, 21)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(40, 25)
        d2.add_gas(0, 21)
        i2 = d2.calculate()

        assert len(i1.stops) > 0
        assert len(i2.stops) > 0

        self._verify(i1, i2, 9)


    def test_oc_ean(self):
        """
        Test shallower OC dive on EAN 32
        """
        d1 = DecoEngine()
        d1.add_segment(32, 30)
        d1.add_gas(0, 21)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(28, 30)
        d2.add_gas(0, 21)
        i2 = d2.calculate()

        assert len(i1.stops) > 0
        assert len(i2.stops) > 0

        self._verify(i1, i2, 9)


    def test_oc_tx(self):
        """
        Test shallower OC dive on TX 18/45
        """
        d1 = DecoEngine()
        d1.add_segment(32, 30)
        d1.add_gas(0, 18, 45)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(28, 30)
        d2.add_gas(0, 18, 45)
        i2 = d2.calculate()

        assert len(i1.stops) > 0
        assert len(i2.stops) > 0

        self._verify(i1, i2, 12)


    def test_ccr_air(self):
        """
        Test shallower CCR dive on Air
        """
        d1 = DecoEngine()
        d1.add_segment(35, 50)
        d1.add_gas(0, 21)
        d1.add_setpoint(0, 1.3)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(32, 50)
        d2.add_gas(0, 21)
        d2.add_setpoint(0, 1.3)
        i2 = d2.calculate()

        assert len(i1.stops) > 0
        assert len(i2.stops) > 0

        self._verify(i1, i2, 9)


    def test_ccr_ean(self):
        """
        Test shallower CCR dive on EAN 32
        """
        d1 = DecoEngine()
        d1.add_segment(35, 50)
        d1.add_gas(0, 32)
        d1.add_setpoint(0, 1.3)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(32, 50)
        d2.add_gas(0, 32)
        d2.add_setpoint(0, 1.3)
        i2 = d2.calculate()

        assert len(i1.stops) > 0
        assert len(i2.stops) > 0

        self._verify(i1, i2, 9)


    def test_ccr_tx(self):
        """
        Test shallower CCR dive on TX 18/45
        """
        d1 = DecoEngine()
        d1.add_segment(35, 50)
        d1.add_gas(0, 18, 45)
        d1.add_setpoint(0, 1.3)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(32, 50)
        d2.add_gas(0, 18, 45)
        d2.add_setpoint(0, 1.3)
        i2 = d2.calculate()

        assert len(i1.stops) > 0
        assert len(i2.stops) > 0

        self._verify(i1, i2, 12)


class ShallowerNDLDiveTestCase(TestCase):
    """
    Shallower depth sanity unit tests for NDL dives.
    """
    def test_oc_ndl_air(self):
        """
        Test shallower OC NDL dive on Air
        """
        d1 = DecoEngine()
        d1.add_segment(20, 25)
        d1.add_gas(0, 21)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(18, 25)
        d2.add_gas(0, 21)
        i2 = d2.calculate()

        assert len(i1.stops) == 0
        assert len(i2.stops) == 0

        self.assertTrue(i1.ndl < i2.ndl)


    def test_oc_ndl_ean(self):
        """
        Test shallower OC NDL dive on EAN 32
        """
        d1 = DecoEngine()
        d1.add_segment(20, 25)
        d1.add_gas(0, 32)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(18, 25)
        d2.add_gas(0, 32)
        i2 = d2.calculate()

        assert len(i1.stops) == 0
        assert len(i2.stops) == 0

        self.assertTrue(i1.ndl < i2.ndl)


    def test_oc_ndl_tx(self):
        """
        Test shallower OC NDL dive on TX 18/45
        """
        d1 = DecoEngine()
        d1.add_segment(20, 25)
        d1.add_gas(0, 18, 45)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(18, 25)
        d2.add_gas(0, 18, 45)
        i2 = d2.calculate()

        assert len(i1.stops) == 0
        assert len(i2.stops) == 0

        self.assertTrue(i1.ndl < i2.ndl)


    def test_ccr_ndl_air(self):
        """
        Test shallower CCR NDL dive on Air
        """
        d1 = DecoEngine()
        d1.add_segment(20, 25)
        d1.add_gas(0, 21)
        d1.add_setpoint(0, 1.3)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(18, 25)
        d2.add_gas(0, 21)
        d2.add_setpoint(0, 1.3)
        i2 = d2.calculate()

        assert len(i1.stops) == 0
        assert len(i2.stops) == 0

        self.assertTrue(i1.ndl < i2.ndl)


    def test_ccr_ndl_ean(self):
        """
        Test shallower CCR NDL dive on EAN 32
        """
        d1 = DecoEngine()
        d1.add_segment(20, 25)
        d1.add_gas(0, 32)
        d1.add_setpoint(0, 1.3)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(18, 25)
        d2.add_gas(0, 32)
        d2.add_setpoint(0, 1.3)
        i2 = d2.calculate()

        assert len(i1.stops) == 0
        assert len(i2.stops) == 0

        self.assertTrue(i1.ndl < i2.ndl)


    def test_ccr_ndl_tx(self):
        """
        Test shallower CCR NDL dive on TX 18/45
        """
        d1 = DecoEngine()
        d1.add_segment(20, 25)
        d1.add_gas(0, 18, 45)
        d1.add_setpoint(0, 1.3)
        i1 = d1.calculate()

        d2 = DecoEngine()
        d2.add_segment(18, 25)
        d2.add_gas(0, 18, 45)
        d2.add_setpoint(0, 1.3)
        i2 = d2.calculate()

        assert len(i1.stops) == 0
        assert len(i2.stops) == 0

        self.assertTrue(i1.ndl < i2.ndl)


# vim: sw=4:et:ai
